import h5py
import numpy as np
import tensorflow as tf


class HDF5Generator:
    def __init__(self, path, prefix, verbose=1):
        self.path = path
        self.prefix = prefix
        self.data = h5py.File(self.path, 'r')
        f = self.data

        self.n_data = f["{}_data".format(self.prefix)].shape[0]
        label_name= "{}_label".format(self.prefix)

        self.n_class = len(np.unique(f[label_name][()]))
        hist = np.histogram(f[label_name][()], bins=self.n_class - 1)
        self.class_weight = (1 / hist[0]) * f[label_name].shape[0] / self.n_class

        if verbose > 0:
            print("=+*"*20)
            print("Dataset {} with {} data".format(path, self.n_data))
            print("=+*"*20)

    def __call__(self):
        with h5py.File(self.path, 'r') as f:
            for d, l in zip(f["{}_data".format(self.prefix)], f["{}_label".format(self.prefix)]):
                yield (d, l)

    def get_dataset(self, input_shape=(300, 30), batch_size=32, shuffle=False, n_shuffle=10000):
        d_set = tf.data.Dataset.from_generator(
            self,
            (tf.float32, tf.int8),
            (tf.TensorShape(input_shape), tf.TensorShape([]))
        )
        d_set = d_set.shuffle(n_shuffle, reshuffle_each_iteration=True) if shuffle else d_set

        return d_set.batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE).repeat()



import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import io


class ConfuseCallback(tf.keras.callbacks.Callback):
    def __init__(self, test_data, test_labels, classes, file_writer, figure_size=(12, 10)):
        super(ConfuseCallback, self).__init__()
        self.test_data = test_data
        self.test_labels = test_labels
        self.classes = classes
        self.file_writer = file_writer
        self.figure_size = figure_size

    def get_precision_recall_plot(self, con_mat):
        precisions = np.array([0] * len(self.classes)).astype('float32')
        recalls = np.array([0] * len(self.classes)).astype('float32')

        for i in range(con_mat.shape[0]):
            tp = con_mat[i, i]
            fn = (con_mat[i, :].sum() - tp)

            fp = (con_mat[:, i].sum() - tp)
            tn = (con_mat.diagonal().sum() - tp)

            # tpr = tp / np.sum(self.test_labels[()] == i)
            # fnr = fn / np.sum(self.test_labels[()] == i)
            # fpr = fp / np.sum(self.test_labels[()] != i)
            # tnr = tn / np.sum(self.test_labels[()] != i)

            precisions[i] = max(0, tp / (tp + fp))
            recalls[i] = max(0, tp / (tp + fn))

        df = pd.DataFrame((self.classes, precisions, recalls)).T
        df.columns = ["Class", "Precision", "Recall"]
        df = pd.melt(df, id_vars="Class", var_name="Type", value_name="Value")

        figure = plt.figure(figsize=self.figure_size)
        sns.barplot(y='Class', x='Value', hue='Type', data=df)
        plt.tight_layout()

        buf = io.BytesIO()
        plt.savefig(buf, format='png')
        plt.close(figure)

        image = tf.image.decode_png(buf.getvalue(), channels=4)
        image = tf.expand_dims(image, 0)

        return image, precisions, recalls

    def on_epoch_end(self, epoch, logs=None):
        try:
            test_pred = self.model.predict(self.test_data)
            test_pred = tf.argmax(test_pred, axis=1)
            accuracy = np.sum(test_pred == self.test_labels) / self.test_labels.shape[0]
            con_mat = tf.math.confusion_matrix(labels=self.test_labels, predictions=test_pred).numpy()
            con_mat_norm = np.around(con_mat.astype('float') / con_mat.sum(axis=1)[:, np.newaxis], decimals=2)

            con_mat_df = pd.DataFrame(con_mat_norm,
                                      index=self.classes,
                                      columns=self.classes)

            precision_recall_image, precisions, recalls = self.get_precision_recall_plot(con_mat)

            figure = plt.figure(figsize=self.figure_size)
            sns.heatmap(con_mat_df, annot=True, cmap=plt.cm.Blues)
            plt.ylabel('True label')
            plt.xlabel('Predicted label')
            plt.title("Accuracy : {:.2f}%, Precision : {:.2f}%, Recall : {:.2f}%".format(accuracy*100, precisions.mean()*100, recalls.mean()*100))

            plt.tight_layout()

            buf = io.BytesIO()
            plt.savefig(buf, format='png')

            plt.close(figure)
            buf.seek(0)
            image = tf.image.decode_png(buf.getvalue(), channels=4)

            image = tf.expand_dims(image, 0)

            # Log the confusion matrix as an image summary.
            with self.file_writer.as_default():
                tf.summary.image("Confusion Matrix", image, step=epoch)
                tf.summary.image("Precision and Recall", precision_recall_image, step=epoch)
        except Exception as e:
            print(e)


class ModelSaverCallback(tf.keras.callbacks.Callback):
    def __init__(self, best_loss=float('inf'), save_root="./", enable=True, epoch=0):
        super(ModelSaverCallback, self).__init__()
        self.best_loss = best_loss
        self.epoch = epoch
        self.save_root = save_root
        self.enable = enable

    def on_epoch_end(self, epoch, logs=None):
        try:
            epoch += self.epoch
            if logs['val_loss'] < self.best_loss:
                # TODO Delete previous saved model if exists
                file_name = '{}my_model_weight_{:04d}_{:03.2f}.h5'.format(self.save_root, epoch, logs['val_loss'])
                print("\nBest loss! saving the model to {} ...".format(file_name))
                self.best_loss = logs['val_loss']

                if self.enable:
                    self.model.save(file_name)
        except Exception as e:
            print(e)


def run_tensorboard(path, port=6006):
    tb = program.TensorBoard()
    tb.configure(argv=[None, '--logdir', path, '--host', '0.0.0.0', '--port', f"{port:}"])
    url = tb.launch()

    print("Running tensorboard on {}".format(url))

    return url


def wait_ctrl_c(pre_msg="Press Ctrl+c to quit Tensorboard", post_msg="\nExit."):
    print(pre_msg)
    try:
        while True:
            time.sleep(3600)
    except:
        print(post_msg)




from moviepy.video.io.ffmpeg_tools import ffmpeg_extract_subclip
import imageio


def generate_segmented_video(vid_info, segment_data, video_root, save_root=None, verbose=0):
    video_name = vid_info['vid_name'][0]
    video_path = video_root + '/' + video_name
    video_name = video_name.replace(".mp4", "")

    if save_root is None:
        save_root = video_root + "/" + "segment/"

    # install ffmpeg plugin
    imageio.plugins.ffmpeg.download()

    video_sync = segment_data['video_sync_time']

    if verbose > 0:
        print("Video sync time :: ", video_sync, "s", end="")

    for i in range(segment_data['segment_x'].shape[0]):
        video_x = segment_data['segment_x'][i]
        segment_name = segment_data['segment_name'][i]
        segment_label = segment_data['segment_label'][i]

        start_t = video_x[0] + video_sync
        end_t = video_x[video_x.shape[0] - 1] + video_sync

        file_name = video_name + "_" + str(i) + "_" + str(segment_label) + "_" + segment_name[0] + ".mp4"

        ffmpeg_extract_subclip(video_path, start_t, end_t, save_root + file_name)

        if verbose > 0:
            print(file_name, "::", start_t, "~", end_t)



import setuptools

setuptools.setup(
    name="moclaphar",
    version="0.0.13",
    license='MIT',
    author="Jongkuk Lim",
    author_email="lim.jeikei@gmail.com",
    description="This packages mainly aims to make an easy process for dataset manipulation.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JeiKeiLim/moclaphar",
    packages=setuptools.find_packages(),
    classifiers=[
        # 패키지에 대한 태그
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'scipy>=1.4.1',
        'matplotlib>=3.2.1',
        'seaborn>=0.10.1',
        'h5py>=2.10.0',
        'pandas>=1.0.3',
        'moviepy>=1.0.3',
      ],
)

"""
Dependencies 자동으로 처리되도록 수정 필요
필요 리스트

scipy
  - (numpy)
pandas
matplotlib
seaborn
moviepy
h5py


tensorflow는 사람마다 환경이 다를수 있으니 건너뛰는게 나을듯
생각해보니 위의 목록은 tensorflow 설치 후 확인한 목록임.
가상환경 다시 만들어서 리스트 다시 뽑아야 함

"""
